############################################################################
#                                                                          #
# Imports & Constants                                                      #
#                                                                          #
############################################################################

import argparse
import csv
import gpudb
import sys
import json
from tabulate import tabulate


CSV = "/tmp/data/dc_shape.csv"

TABLE_DC = "dc_shape"

GRAPH_DC = TABLE_DC + "_graph"
SOLUTION_GRAPH_DC_1 = GRAPH_DC + "_solved_sp"
SOLUTION_GRAPH_DC_2 = GRAPH_DC + "_solved_sp_intersection-penalty"
SOLUTION_GRAPH_DC_3 = GRAPH_DC + "_solved_sp_turn-restriction"

############################################################################
#                                                                          #
# Functions                                                                #
#                                                                          #
############################################################################

def table_setup():

    print("\n===========")
    print("TABLE SETUP")
    print("===========\n")

    # Check to see if CSV file exists locally; if it doesn't, exit.
    print("Checking to see if {} file exists locally:".format(CSV))
    try:
        dc_data = csv.reader(open(CSV))
    except IOError as io_error:
        print("- {} \nExiting".format(io_error))
        sys.exit()
    print("- {} file exists!\n".format(CSV))

    # Define the TABLE_DC type
    columns = [
        ["link_id", "long", "shard_key"],
        ["direction", "int"],
        ["speed", "double"],
        ["road_type", "string", "char1"],
        ["seg", "string", "wkt"],
        ["shape", "string", "wkt"],
    ]

    # Create the TABLE_DC table
    t = gpudb.GPUdbTable(
        _type=columns,
        name=TABLE_DC,
        db=kinetica,
        options={}
    )

    # Insert records into the table
    print("Creating records from the " + CSV + " file:")
    dc_data.next()  # skip the header
    records = []
    for record in dc_data:
        record_data = [
            int(record[0]),   # link_id
            int(record[1]),   # direction
            float(record[2]), # speed
            record[3],        # road_type
            record[4],        # seg
            record[5]         # shape
        ]
        records.append(record_data)
    print("- {} records created\n".format(len(records)))

    print(
        "Inserting records into {} table. This may take a "
        "while.".format(TABLE_DC)
    )
    t.insert_records(records)
    print("- {} records inserted".format(t.size()))

    # end table_setup()

def graph_setup():

    print("\n===========")
    print("GRAPH SETUP")
    print("===========\n")

    # Create a graph from TABLE_DC
    print("Creating graph: {}".format(GRAPH_DC))
    cg = kinetica.create_graph(
        graph_name=GRAPH_DC,
        directed_graph=True,
        nodes=[],
        edges=[
            TABLE_DC + ".link_id AS EDGE_ID",
            TABLE_DC + ".shape AS EDGE_WKTLINE",
            TABLE_DC + ".direction AS EDGE_DIRECTION",
            "(ST_LENGTH(" + TABLE_DC + ".shape,1)/"
            "(ST_NPOINTS(" + TABLE_DC + ".shape)-1))/" + 
            TABLE_DC + ".speed AS EDGE_WEIGHT_VALUESPECIFIED"
        ],
        weights=[],
        restrictions=[],
        options={
            "recreate": "true",
            "add_turns": "true"
        }
    )
    print("- {}".format(cg["status_info"]["status"]))
    # end graph_setup()

def solve_graph():

    # Set source and destination points for the examples
    src = ["POINT(-77.04489135742188 38.91112899780273)"]
    dest = ["POINT(-77.03193664550781 38.91188049316406)"]

    print("\n====================================")
    print("SOLVE GRAPH - TURN PENALTY EXAMPLE 1")
    print("====================================\n")

    # Solve GRAPH_DC for shortest path between given source_node 'src' and 
    # destination node 'dest'
    sg = kinetica.solve_graph(
        graph_name=GRAPH_DC,
        solver_type="SHORTEST_PATH",
        source_nodes=src,
        destination_nodes=dest,
        solution_table=SOLUTION_GRAPH_DC_1,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"]
    print(
        "Cost for shortest path between source ({}) and "
        "destination ({}):".format(src, dest)
    )

    # Divide the individual destination costs by 60 to convert to minutes
    sg[:] = [
        cost / 60 for cost in sg
    ]

    # Display the results
    headers = ["Destination Node", "Cost (in minutes)"]
    print(
        tabulate(
            {
                headers[0]: dest,
                headers[1]: sg
            },
            headers="keys",
            tablefmt="grid"
        )
    )

    print("\n====================================")
    print("SOLVE GRAPH - TURN PENALTY EXAMPLE 2")
    print("====================================\n")

    # Solve GRAPH_DC for shortest path between given source_node 'src' and
    # destination node 'dest' but with a penalty on using intersections
    sg = kinetica.solve_graph(
        graph_name=GRAPH_DC,
        solver_type="SHORTEST_PATH",
        source_nodes=src,
        destination_nodes=dest,
        solution_table=SOLUTION_GRAPH_DC_2,
        options={
            "export_solve_results": "true",
            "intersection_penalty": "20"
        }
    )["result_per_destination_node"]
    print(
        "Cost for shortest path between source ({}) and "
        "destination ({}) with intersection penalty:".format(src, dest)
    )

    # Divide the individual destination costs by 60 to convert to minutes
    sg[:] = [
        cost / 60 for cost in sg
    ]

    # Display the results
    headers = ["Destination Node", "Cost (in minutes)"]
    print(
        tabulate(
            {
                headers[0]: dest,
                headers[1]: sg
            },
            headers="keys",
            tablefmt="grid"
        )
    )

    print("\n====================================")
    print("SOLVE GRAPH - TURN PENALTY EXAMPLE 3")
    print("====================================\n")

    # Solve GRAPH_DC for shortest path between given source_node 'src' and
    # destination node 'dest' with restricting the turn between given edge IDs
    sg = kinetica.solve_graph(
        graph_name=GRAPH_DC,
        restrictions=[
            "{18352169} AS RESTRICTIONS_FROM_EDGE_ID",
            "{18352166} AS RESTRICTIONS_TO_EDGE_ID",
            "{0} AS RESTRICTIONS_ONOFFCOMPARED"
        ],
        solver_type="SHORTEST_PATH",
        source_nodes=src,
        destination_nodes=dest,
        solution_table=SOLUTION_GRAPH_DC_3,
        options={
            "export_solve_results": "true"
        }
    )["result_per_destination_node"]
    print(
        "Cost for shortest path between source ({}) and "
        "destination ({}) with one turn restriction:".format(src, dest)
    )

    # Divide the individual destination costs by 60 to convert to minutes
    sg[:] = [
        cost / 60 for cost in sg
    ]

    # Display the results
    headers = ["Destination Node", "Cost (in minutes)"]
    print(
        tabulate(
            {
                headers[0]: dest,
                headers[1]: sg
            },
            headers="keys",
            tablefmt="grid"
        )
    )

    # end solve_graph()

def clear_tables():

    # Drop all the tables
    for table_name in [
        TABLE_DC,
        SOLUTION_GRAPH_DC_1,
        SOLUTION_GRAPH_DC_2,
        SOLUTION_GRAPH_DC_3
    ]:
        ct = kinetica.clear_table(table_name, "", {"no_error_if_not_exists": "true"})

    # end clear_tables()

if __name__ == '__main__':

    # Set up args
    parser = argparse.ArgumentParser(description='Run multiple supply & demand match graph example.')
    parser.add_argument('command', nargs="?", help='command to execute (currently only "clear" to remove the example tables')
    parser.add_argument('--host', default='127.0.0.1', help='Kinetica host to run example against')
    parser.add_argument('--username', default='', help='Username of user to run example with')
    parser.add_argument('--password', default='', help='Password of user')

    args = parser.parse_args()

    # Establish connection with a locally-running instance of Kinetica
    kinetica = gpudb.GPUdb(host = ['http://' + args.host + ':9191'], username = args.username, password = args.password)

    # If command line arg is clear, just clear tables and exit
    if (args.command == "clear"):
        clear_tables()
        quit()

    # Execute defined functions
    clear_tables()
    table_setup()
    graph_setup()
    solve_graph()
